"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const chunk_1 = __importDefault(require("lodash/chunk"));
const range_1 = __importDefault(require("lodash/range"));
const perp_1 = require("../../src/accounts/perp");
const client_1 = require("../../src/client");
const constants_1 = require("../../src/constants");
const rpc_1 = require("../../src/utils/rpc");
// Env vars
const CLUSTER = process.env.CLUSTER_OVERRIDE || 'mainnet-beta';
const CLUSTER_URL = process.env.CLUSTER_URL_OVERRIDE || process.env.MB_CLUSTER_URL;
const USER_KEYPAIR = process.env.USER_KEYPAIR_OVERRIDE || process.env.MB_PAYER_KEYPAIR;
const MANGO_ACCOUNT_PK = process.env.MANGO_ACCOUNT_PK || '';
const INTERVAL_UPDATE_BANKS = Number(process.env.INTERVAL_UPDATE_BANKS || 60);
const INTERVAL_CONSUME_EVENTS = Number(process.env.INTERVAL_CONSUME_EVENTS || 5);
const INTERVAL_UPDATE_FUNDING = Number(process.env.INTERVAL_UPDATE_FUNDING || 5);
const INTERVAL_CHECK_NEW_LISTINGS_AND_ABORT = Number(process.env.INTERVAL_CHECK_NEW_LISTINGS_AND_ABORT || 120);
async function updateBanks(client, group) {
    console.log('Starting updateBanks loop');
    // eslint-disable-next-line no-constant-condition
    while (true) {
        const tokenIndices = Array.from(group.banksMapByTokenIndex.keys());
        const tokenIndicesByChunks = (0, chunk_1.default)(tokenIndices, 10);
        tokenIndicesByChunks.map(async (tokenIndices) => {
            const ixs = await Promise.all(tokenIndices.map((ti) => client.tokenUpdateIndexAndRateIx(group, group.getFirstBankByTokenIndex(ti).mint)));
            try {
                const sig = await (0, rpc_1.sendTransaction)(client.program.provider, ixs, group.addressLookupTablesList, { prioritizationFee: true });
                console.log(` - Token update index and rate success, tokenIndices - ${tokenIndices}, sig https://explorer.solana.com/tx/${sig.signature}`);
            }
            catch (e) {
                console.log(` - Token update index and rate error, tokenIndices - ${tokenIndices}, e - ${e}`);
            }
        });
        await new Promise((r) => setTimeout(r, INTERVAL_UPDATE_BANKS * 1000));
    }
}
async function consumeEvents(client, group) {
    console.log('Starting consumeEvents loop');
    // eslint-disable-next-line no-constant-condition
    while (true) {
        const perpMarketIndices = Array.from(group.perpMarketsMapByMarketIndex.keys());
        for (const perpMarketIndex of perpMarketIndices) {
            for (const unused of (0, range_1.default)(0, 10)) {
                const perpMarket = group.getPerpMarketByMarketIndex(perpMarketIndex);
                const eq = await perpMarket.loadEventQueue(client);
                const events = eq.getUnconsumedEvents().slice(0, 10);
                const accounts = new Set();
                for (const event of events) {
                    if (event.eventType === perp_1.PerpEventQueue.FILL_EVENT_TYPE) {
                        accounts.add(event.maker);
                        accounts.add(event.taker);
                    }
                    else if (event.eventType === perp_1.PerpEventQueue.OUT_EVENT_TYPE) {
                        accounts.add(event.owner);
                    }
                    else if (event.eventType === perp_1.PerpEventQueue.LIQUIDATE_EVENT_TYPE) {
                        // pass
                    }
                }
                try {
                    const sig = await (0, rpc_1.sendTransaction)(client.program.provider, [
                        await client.perpConsumeEventsIx(group, perpMarketIndex, Array.from(accounts), 10),
                    ], group.addressLookupTablesList, { prioritizationFee: true });
                    console.log(` - Consume events success, perpMarketIndex - ${perpMarketIndex}, sig https://explorer.solana.com/tx/${sig.signature}`);
                }
                catch (e) {
                    console.log(` - Consume events error, perpMarketIndex - ${perpMarketIndex}, e - ${e}`);
                }
            }
        }
        await new Promise((r) => setTimeout(r, INTERVAL_CONSUME_EVENTS * 1000));
    }
}
async function updateFunding(client, group) {
    console.log('Starting updateFunding loop');
    // eslint-disable-next-line no-constant-condition
    while (true) {
        const perpMarketIndices = Array.from(group.perpMarketsMapByMarketIndex.keys());
        for (const perpMarketIndex of perpMarketIndices) {
            try {
                const status = await (0, rpc_1.sendTransaction)(client.program.provider, [
                    await client.perpUpdateFundingIx(group, group.getPerpMarketByMarketIndex(perpMarketIndex)),
                ], group.addressLookupTablesList, { prioritizationFee: true });
                console.log(` - Update funding success, perpMarketIndex - ${perpMarketIndex}, sig https://explorer.solana.com/tx/${status.signature}`);
            }
            catch (e) {
                console.log(` - Update funding error, perpMarketIndex - ${perpMarketIndex}, e - ${e}`);
            }
        }
        await new Promise((r) => setTimeout(r, INTERVAL_UPDATE_FUNDING * 1000));
    }
}
async function checkNewListingsAndAbort(client, group) {
    console.log('Starting checkNewListingsAndAbort loop');
    // eslint-disable-next-line no-constant-condition
    while (true) {
        const freshlyFetchedGroup = await client.getGroup(group.publicKey);
        if (freshlyFetchedGroup.banksMapByTokenIndex.size !=
            group.banksMapByTokenIndex.size ||
            freshlyFetchedGroup.perpMarketsMapByMarketIndex.size !=
                group.perpMarketsMapByMarketIndex.size) {
            process.exit();
        }
        await new Promise((r) => setTimeout(r, INTERVAL_CHECK_NEW_LISTINGS_AND_ABORT * 1000));
    }
}
async function keeper() {
    // Load client
    const options = anchor_1.AnchorProvider.defaultOptions();
    const connection = new web3_js_1.Connection(CLUSTER_URL, options);
    const user = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(process.env.KEYPAIR || fs_1.default.readFileSync(USER_KEYPAIR, 'utf-8'))));
    const userWallet = new anchor_1.Wallet(user);
    const userProvider = new anchor_1.AnchorProvider(connection, userWallet, options);
    const client = await client_1.MangoClient.connect(userProvider, CLUSTER, constants_1.MANGO_V4_ID[CLUSTER], {
        idsSource: 'get-program-accounts',
    });
    const mangoAccount = await client.getMangoAccount(new web3_js_1.PublicKey(MANGO_ACCOUNT_PK));
    const group = await client.getGroup(mangoAccount.group);
    await group.reloadAll(client);
    updateBanks(client, group);
    consumeEvents(client, group);
    updateFunding(client, group);
    checkNewListingsAndAbort(client, group);
}
keeper();
